package bogue

  1. Overview
  2. Docs

Module Bogue.MouseSource

Mouse and touchscreen information.

All positions are given in logical pixels, not hardware pixels (see the Theme SCALE variable).

Dependency graph
Sourceval pointer_pos : Tsdl.Sdl.event -> int * int

Return the pointer current position. This works both for mouse and touchscreen.

Sourceval pos : unit -> int * int

Get current mouse position.

Sourceval window_pos : Tsdl.Sdl.window -> int * int

Return the mouse position with respect to the given window, using window position if necessary.

Sourceval button_pos : Tsdl.Sdl.event -> int * int
Sourceval finger_pos : Tsdl.Sdl.event -> float * float