Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Page
Library
Module
Module type
Parameter
Class
Class type
All notable changes to the Bogue project will be documented in this file.
The format is based on Keep a Changelog.
Ti_Validator (automatic validating text inputs). Includes an email validator.Mailbox for sending arbitrary message to widgets (experimental, but should be quite powerful to help organizing large and complex UIs)I18n.make_context is now public.breaking! Colors are now separated in two modules RGB and RGBA. Several functions have been moved to these modules. If you want to cheaply workaround this, just add
open Bogue.RGBA
open Bogue.RGBat the top of your file, this should do the job in most cases.
max_size of Text_input.create and Widget.text_input is now renamed to max_len ("size" is used for font size).colors dir from the common themes dir.BOX_BG_COLORI18n (Internationalization), for translating string to the user's language. Configuration files for 15 languages.[DISABLED_BG] and [DISABLED_FG].Timeout.add_ignore.Widget.remove_connection, Widget.remove_trigger.Popup.one_button and Popup.two_buttons now have a bg optional argument.File: fix file dialog crashing when monitoring a directory being deleted.File for file selection dialogs and file monitoring. This is quite a complex layout, so we decided make it public now in a beta-testing phase. New features will be added later, and maybe small API changes.Popups can now be closed by pressing ESCAPE.Main for dynamically updating global shortcuts.Table.create now returns a private type. In order to obtain the layout you should then apply Table.get_layout.Layout.set_size now takes optional arguments for sizes (allows to reduce the number of calls to resize_content)Theme.set_integer_scale is renamed to Theme.set_int_scale to conform with the env var INT_SCALE.Mouse.wheel_y, Mouse.set_wheel_speed.Flat and Tower have new resizing strategies.Selection new functions is_empty, size, range, first, last, invert.Table layouts can now be resized dynamically.tabs with TAB! ;) (preliminary)Layout.set_enabled. New Examples #53, #53bis.Text_display is now possible, using Widget.html, see exemple 47.Text_display is now possible, using Widget.html (but spaces between words are not underlined at this point), see example 6.Theme.get_path is now public.Theme.get_font_path_optLabel.font_from_file allows the use of arbitrary fonts in addition to the Theme fonts (which is not recommended anyways for good design).Layout.flat* and Layout.tower* have changed. If you were using any of scale_content or keep_resize you should use the new resize option instead.Layout.unload_textures is renamed Layout.iter_unload_texturesTable layouts can now be resized dynamically.tabs: menu height is fixed.flat and tower: margins are not scaled.Long_list layouts can now be resized dynamically.Label style parameter is now usable (ocamldoc invocation was wrong for tsdl_ttf), and can also be used for Widget.label.Resize module inside Layout.dst field of blit is None means NO RENDERING.Detect for detecting flat and tower layouts.Avar.reset is now public.Layout.is_shown.Layout.inside is now public.Snapshot.to_cursor (experimental).Main.get_monitors_refresh_rate.Main.run and Time.adaptive_fps have an optional argument vsync.sdl_event and Sdl.Event.enum.Time.adaptive_fps using swap interval.Main.get_window_refresh_rateDraw.surface_from_textureText_input.activate is now publicBox.get_style and Box.set_style are now public.mouse_at_rest now uses Timeout.directories to xdg to work on Windows.example.ml file is now called examples.ml, which is much more logical since it contains more than 50 examples.module Time = B_time.directories to compute home and config paths, and add special care for Windows OS.